# include	<stdio.h>
# include	<stdlib.h>

# include	"vcrplus.h"
# include	"ileave.h"
# include	"misc.h"
# include	"scramble.h"
# include	"b3.h"
# include	"top456.h"

long		debug = 0;


static char *mon[] = {  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };


long
decode( long month_today, long date_today, long year_today, long newspaper )
{
	long	s1_out, bot3, top5, quo, rem ;
	long	s5_out, mtout, ofout, tval, cval ;
	long	day_out, channel_out;
	long	starttime_out, duration_out ;

	year_today = year_today % 100 ;
	debug = 0;

	if ((month_today > 12) || (month_today < 1) ||
	      (date_today < 1) || (date_today > 31)) {
		fputs("Invalid date\n", stdout);
	}

	if ((newspaper < 1) || (newspaper > 999999)) {
		fputs("Sorry, I can't process that code value\n", stdout);
		fputs("	I can only do codes with 6 digits or fewer\n", stdout);
	}

	ofout = INVALID ;
	mtout = INVALID ;

	s1_out = f1(newspaper, KEY001); 
	bot3 = s1_out % 1000;
	top5 = s1_out / 1000;
	quo = (bot3 - 1) / 32 ;
	rem = (bot3 - 1) % 32 ;
	day_out = quo + 1;

	/* spot instances of next month's/year's codes - S.Hosgood, 8.Feb.93 */
	if (day_out < date_today) {
		month_today++;

		if (month_today > 12) {
			month_today = 1;
			year_today = (year_today+1) % 100;
		}
	}

	if (newspaper >= 1000)
		offset(day_out, year_today, top5, (&ofout), (&mtout)) ;
	else {
		mtout = 0;
		ofout = 0;
	}

	s5_out = (rem + (day_out*(month_today+1)) + ofout) % 32 ;


	de_interleave(mtout, s5_out, &tval, &cval);
	channel_out = cval + 1;
	lookup(tval, (&starttime_out), (&duration_out));

	printf("\t%s  %2ld  %2ld\n\tChannel= %2ld\n", 
		mon[month_today-1], day_out, year_today, channel_out);
	printf("\tStart_time= %04ld  Length= %3ld\n",
		starttime_out, duration_out);
	printf("\tEnd_time= %04ld\n",
		end_time(starttime_out, duration_out));

	return tval;
}

int
channel( long month_today, long date_today, long year_today, long newspaper )
{
	long	s1_out, bot3, top5, quo, rem ;
	long	s5_out, mtout, ofout, tval, cval ;
	long	day_out, channel_out;
	long	starttime_out, duration_out ;

	year_today = year_today % 100 ;
	debug = 0;

	if( (month_today>12) || (month_today<1) || (date_today<1) || (date_today>31) ) {
		fputs("Invalid date\n", stdout);
	}

	if((newspaper < 1) || (newspaper > 999999)) {
		fputs("Sorry, I can't process that code value\n", stdout);
		fputs("	I can only do codes with 6 digits or fewer\n", stdout);
	}

	ofout = INVALID ;
	mtout = INVALID ;

	s1_out = f1(newspaper, KEY001); 
	bot3 = s1_out % 1000;
	top5 = s1_out / 1000;
	decode_bot3(bot3, &day_out, &rem);

	/* spot instances of next month's/year's codes - S.Hosgood, 8.Feb.93 */
	if (day_out < date_today)
		month_today++;
	if (month_today > 12) {
		month_today = 1;
		year_today = (year_today+1) % 100;
	}

	if(newspaper >= 1000)
		offset(day_out, year_today, top5, (&ofout), (&mtout)) ;
	else {
		mtout = 0;
		ofout = 0;
	}

	s5_out = (rem + (day_out*(month_today+1)) + ofout) % 32 ;


	de_interleave(mtout, s5_out, &tval, &cval);
	return cval + 1;
}


static void
usage( char *progname )
{

# ifdef US
	printf("Usage: %s <Month (1-12)> <Day (1-31)> <Year (0-99)>\n",
								progname);
# endif
# ifdef UK
	printf("Usage: %s <Day (1-31)> <Month (1-12)> <Year (0-99)>\n",
								progname);
# endif
	return;
}


int
main(int argc, char *argv[])
{
	char		buffer[20];
	int		loop = 1;
	long		n;
	long		month_today, date_today, year_today, newspaper;

	if (argc == 4) {
# ifdef US
		month_today = atol(argv[1]);
		date_today = atol(argv[2]);
# endif
# ifdef UK
		date_today = atol(argv[1]);
		month_today = atol(argv[2]);
# endif
		year_today = atol(argv[3]);
	}
	else {
		usage("interact");
		return -1;
	}

	while (loop) {
		fputs("Code? ", stdout);
		fgets(buffer, sizeof buffer, stdin);

		switch (buffer[0]) {
		case 'c':
		case 'C':
			/* match this channel-no */
			/* experiment to find that elusive factor! */
			n = atol(buffer+1);

			for (magic = 1; magic < 50; magic++) {
				if (channel(month_today, date_today, year_today, newspaper) == n) {
					printf("Magic = %d, ", magic);
				}
			}

			break;

		case 'm':
		case 'M':
			/* set magic number */
			magic = atol(buffer+1);
			decode(month_today, date_today, year_today, newspaper);
			break;

		case 'q':
		case 'Q':
			loop = 0;
			break;

		default:
			newspaper = atol(buffer);
			n = decode(month_today, date_today, year_today, newspaper);

			/* if it was a guess, allow him to confirm */
			confirm(n);
		}

		putchar('\n');
	}

	/* maybe write table (if it changed) */
	dump();
	return 0;
}
