# include	<stdio.h>
# include	<stdlib.h>
# include	<malloc.h>
# ifdef MSDOS
# include	<conio.h>
# endif /* MSDOS */
# include	<string.h>
# include	<errno.h>

# include	"misc.h"
# include	"b3.h"
# include	"fnc7.h"
# include	"ileave.h"
# include	"scramble.h"
# include	"lookup.h"

long	debug = 0;

struct code {
	int		serial_no;
	long		code;
	int		index;
	struct code	*next;
	struct code	*chain;
};

struct group {
	int		serial_no;
	int		type;
# define	HHB		1
# define	QHB		2
# define	PATTERN		3
# define	ERRATIC		4
	int		id;
	int		len;
	long		tblindex;
	struct group	*nextgrp;
	struct group	*lastgrp;
	struct group	*nextlist;
	struct group	*lastlist;
	struct code	*ch1to4;
	struct code	*rest;
};

int		group_count = 0, code_count = 0;
struct group	*master_list = NULL;
int		known, total, terrestial;


void estimates( void );
void read_table(char *filename);
int spantest( struct group *gp, struct code *cp, int *tally );
struct group *newgroup( int type, int id, int len, long tblindex );
struct code *newcode(struct code *last,struct code *next,int index,long code);
struct group *insert( int type, int id, int len, int index, int channel,
				long code, struct group *prev, long tblindex );
void write_table( void );
char *fgetl( char *buffer, int bufsiz, FILE *in );
struct group *parse_reply( long top5, int first_reqd, struct group *prev,
					int *np );

int
main( int argc, char *argv[] )
{
	static char	buffer[32];
	int		bot3, rem, s5_out, day, n;
	long		newspaper, top5, s4_out, start, stop;
	struct group	*prev;

	if (argc > 3) {
		fprintf(stderr, "Usage:\tlist7 <start-code> <stop-code>\n");
		fprintf(stderr, "or:\tlist7 <individual group to test>\n");
		fprintf(stderr, "or:\tlist7 [<number of tests>] (default=10)\n");
		return -1;
	}

	read_table("list7.tbl");

	switch (argc) {
	case 1:
		start = 1000 + total;
		stop = start + 10;
		break;

	case 2:
		if (atol(argv[1]) > 999)
			start = stop = atol(argv[1]);
		else {
			start = 1000 + total;
			stop = start + atol(argv[1]) - 1;
		}

		break;

	case 3:
		start = atol(argv[1]);
		stop = atol(argv[2]);
	}

	printf("Estimates (as of last time):\n");
	estimates();
	fprintf(stderr, "Target date? ");
	fgets(buffer, sizeof buffer, stdin);
	day = atoi(buffer);

	for (top5 = start; top5 <= stop; top5++) {
		/* we only want s5_out really */
		/* do this for the first of the group-of-eight */
		interleave(0L, 0, (&s4_out), (&s5_out));
		if ((rem = s5_out - makebump(top5)) < 0)
			rem += 32;

		bot3 = encode_bot3(day, rem);
		newspaper = top5*1000 + bot3;
		newspaper = f1(newspaper, KEY002);
		printf("Group %ld: Decode %ld: ", top5, newspaper);
		prev = parse_reply(top5, 0, NULL, &n);

		if (prev != NULL) {
			/* do this again for the one after the link */
			interleave(( long ) n, 0, (&s4_out), (&s5_out));
			if ((rem = s5_out - makebump(top5)) < 0)
				rem += 32;

			bot3 = encode_bot3(day, rem);
			newspaper = top5*1000 + bot3;
			newspaper = f1(newspaper, KEY002);
			printf("Also interested in Group %ld+%d: Decode %ld: ",
					top5, n, newspaper);

			parse_reply(top5, 1, prev, &n);
		}

		total++;
	}

	write_table();
	printf("Current estimates:\n");
	estimates();
	return 0;
}

void
estimates( void )
{

	/* the 2000 is the size of the table below 7-digit codes */
	printf("7 fig table size = %.1f\n",
		2000.0*(((( double ) total) / ( double ) known) - 1.0));

	/* the 4 is the number of channels per "clump" */
	printf("Total number of channels in 7 fig scope = %.1f\n",
		4.0*((( double ) total) / ( double ) terrestial));
	return;
}

void
read_table(char *filename)
{
	static char	buffer[120], *cptr;
	int		n, i, size = 0;
	int		itotal, iknown = 0, iterrestial = 0;
	int		lstart, llen = 0;
	FILE		*in;
	struct group	**gblock, *gp, *tp;
	struct code	**cblock, *cp;
# ifdef FIND_MISSING
	static char	missing[200];

	for (i = 0; i < 200; i++)
		missing[i] = ( char ) 0;
# endif
	if ((in = fopen(filename, "r")) != NULL) {
		/* first two lines of file are numbers of structs reqd */
		fgetl(buffer, sizeof buffer, in);
		group_count = atoi(buffer);
		fgetl(buffer, sizeof buffer, in);
		itotal = code_count = atoi(buffer);

		/* dynamically allocate tracking-arrays */
		gblock = ( struct group ** ) malloc(sizeof( struct group * ) * group_count);
		cblock = ( struct code ** ) malloc(sizeof( struct code * ) * code_count);

		/*
		 * allocate correct numbers of structs and
		 * keep arrays of pointers to them
		 */
		for (n = 0; n < group_count; n++)
			gblock[n] = ( struct group * ) malloc(sizeof( struct group ));
		for (n = 0; n < code_count; n++)
			cblock[n] = ( struct code * ) malloc(sizeof( struct code ));

		/* 3rd line of file is index to head of master-list */
		fgetl(buffer, sizeof buffer, in);
		master_list = ((i = atoi(buffer)) != -1)? gblock[i]: NULL;

		/* next 3 lines contain various tallies */
		fgetl(buffer, sizeof buffer, in);
		known = atoi(buffer);	/* number of known bunches-of-8 */
		fgetl(buffer, sizeof buffer, in);
		total = atoi(buffer);	/* total number of bunches */
		fgetl(buffer, sizeof buffer, in);
		terrestial = atoi(buffer); /* number of bunches on ch1-4 */

		/*
		 * fill structs from datafile -
		 * one comma-separated line per struct.
		 */
		for (n = 0; n < group_count; n++) {
			/* four integers, a long, four group ptrs, two code ptrs */
			/* NULL ptrs are stored as -1 */
			fgetl(buffer, sizeof buffer, in);

			if ((i = atoi(buffer)) < 0 || i >= group_count) {
				fprintf(stderr, "Illegal group serial-no %d\n", i);
				exit(-1);
			}

			gp = gblock[i];
			gp->serial_no = i;
			cptr = strchr(buffer, ',');
			if (cptr == NULL)
				break;
			gp->type = atoi(++cptr);
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->id = atoi(++cptr);
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->len = atoi(++cptr);
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->tblindex = atol(++cptr);
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->nextgrp = ((i = atoi(++cptr)) != -1)? gblock[i]: NULL;
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->lastgrp = ((i = atoi(++cptr)) != -1)? gblock[i]: NULL;
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->nextlist = ((i = atoi(++cptr)) != -1)? gblock[i]: NULL;
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->lastlist = ((i = atoi(++cptr)) != -1)? gblock[i]: NULL;
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->ch1to4 = ((i = atoi(++cptr)) != -1)? cblock[i]: NULL;
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			gp->rest = ((i = atoi(++cptr)) != -1)? cblock[i]: NULL;

			if (gp->tblindex == -1) {
				/* total number of 7-fig-only timeslots */
				switch (gp->type) {
				case HHB:
				case QHB:
					size += 48;
					break;

				case PATTERN:
					size += 34;
				}
			}
		}

		for (n = 0; n < code_count; n++) {
			/* three integers, one code ptr */
			/* NULL ptrs are stored as -1 */
			fgetl(buffer, sizeof buffer, in);
			if ((i = atoi(buffer)) < 0 || i >= code_count) {
				fprintf(stderr, "Illegal code serial-no %d\n", i);
				exit(-1);
			}

			cp = cblock[i];
			cp->serial_no = i;
			cptr = strchr(buffer, ',');
			if (cptr == NULL)
				break;
			cp->code = atol(++cptr);
# ifdef FIND_MISSING
			missing[( int )(cp->code) - 1000] = ( char ) 1;
# endif
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			cp->index = atoi(++cptr);
			cptr = strchr(cptr, ',');
			if (cptr == NULL)
				break;
			cp->next = ((i = atoi(++cptr)) != -1)? cblock[i]: NULL;
		}

		/* find groups whose final code spans to another group */
		for (n = 0; n < group_count; n++) {
			gp = gblock[n];

			if (gp->tblindex != -1) {
				if (gp->ch1to4 != 0)
					fprintf(stderr, "\agblock[%d] has 6-digit code!\n", n);

			/*	itotal -= spantest(gp, gp->ch1to4, NULL); */
				itotal -= spantest(gp, gp->rest, &iknown);
			}
			else {
				itotal -= spantest(gp, gp->ch1to4, &iterrestial);
				itotal -= spantest(gp, gp->rest, NULL);
			}
		}

		/* search for longest chain */
		for (tp = master_list; tp; tp = tp->nextlist) {
			if (tp->tblindex != -1)
				continue;

			for (gp = tp, i = 0; gp; gp = gp->nextgrp, i++);

			if (i > llen) {
				llen = i;
				lstart = tp->serial_no;
			}
		}

		printf("Longest chain = %d groups from %d\n", llen, lstart);

# ifdef FIND_MISSING
		for (n = 0; n <= itotal; n++) {
			if (missing[n] == ( char ) 0)
				printf("Redo: %d\n", n+1000);
		}
# endif
		/* free dynamically allocated tracking-arrays */
		free(( void * ) gblock);
		free(( void * ) cblock);
		fclose(in);

		printf("%d timeslots accounted-for so far\n", size);
		printf("File says %d: I make it %d known code bunches\n", known, iknown);
		printf("File says %d: I make it %d total code bunches\n", total, itotal);
		printf("File says %d: I make it %d terrestial code bunches\n", terrestial, iterrestial);
		total = itotal;
		known = iknown;
		terrestial = iterrestial;
	}

	return;
}


int
spantest( struct group *gp, struct code *cp, int *tally )
{
	int	itotal = 0;

	while (cp && cp->next) {
		if (tally != NULL)
			(*tally)++;

		cp = cp->next;
	}

	if (cp) {
		switch (gp->type) {
		case HHB:
		case QHB:
			if (cp->index > 48 - 8) {
				/* spans */
				itotal++;

				if (gp->nextgrp == NULL)
					printf("Fix %ld\n", cp->code);
			}
			else if (tally != NULL)
				/* no span, and we're tallying */
				(*tally)++;

			break;

		case PATTERN:
			if (cp->index > 34 - 8) {
				/* spans */
				itotal++;

				if (gp->nextgrp == NULL)
					printf("Fix %ld\n", cp->code);
			}
			else if (tally != NULL)
				/* no span, and we're tallying */
				(*tally)++;
		}
	}

	return itotal;
}


struct group *
insert( int type, int id, int len, int index, int channel, long code, struct group *prev, long tblindex )
{
	struct group	*gp, *tp;
	struct code	*cp, *lastcp;

	for (tp = master_list; tp; tp = tp->nextlist) {
		for (gp = tp; gp; gp = gp->nextgrp) {
			if (gp->type != type) continue;
			if (gp->id != id) continue;
			if (gp->len != len) continue;

			if (gp->tblindex != tblindex)
				fprintf(stderr, "\aTable-indexes don't tie up!\n");

			/* this seems to be the one */
			cp = (channel != -1)? gp->ch1to4: gp->rest;
			lastcp = NULL;

			while (cp) {
				if (cp->index == index) {
					if (cp->code == code) {
						printf("\aDuplicate?\n");
						break;
					}

					cp = newcode(lastcp, cp, index, code);
					if (lastcp == NULL) {
						if (channel != -1)
							gp->ch1to4 = cp;
						else 
							gp->rest = cp;
					}

					break;
				}
				else if (cp->index > index) {
					cp = newcode(lastcp, cp, index, code);

					if (lastcp == NULL) {
						if (channel != -1)
							gp->ch1to4 = cp;
						else 
							gp->rest = cp;
					}

					break;
				}

				lastcp = cp;
				cp = cp->next;
			}

			if (cp == NULL) {
				/* insert at end of list then */
				cp = newcode(lastcp, NULL, index, code);

				if (lastcp == NULL) {
					if (channel != -1)
						gp->ch1to4 = cp;
					else 
						gp->rest = cp;
				}
			}

			if (prev) {
				if (prev->nextgrp) {
					if (prev->nextgrp != gp)
						printf("\aPrevious group already tied!!\n");
					else
						printf("\aPrevious group already tied to here!!\n");

					return NULL;
				}

				if (gp->lastgrp) {
					printf("\aThis group already tied!!\n");
					return NULL;
				}

				/* check gp isn't the head of the same chain as prev */
				for (tp = prev; tp; tp = tp->lastgrp);

				if (tp == gp) {
					printf("\aCan't self-link!!\n");
					return NULL;
				}

				/* unlink gp from master-list */
				if (gp->lastlist)
					gp->lastlist->nextlist = gp->nextlist;
				else
					master_list = gp->nextlist;

				if (gp->nextlist)
					gp->nextlist->lastlist = gp->lastlist;

				/* link gp after prev */
				prev->nextgrp = gp;
				gp->lastgrp = prev;
				gp->nextlist = gp->lastlist = NULL;
			}

			return gp;
		}
	}

	/* did not match existing group */
	if ((gp = newgroup(type, id, len, tblindex)) == NULL) {
		fprintf(stderr, "Out of memory!\n");
		return NULL;
	}

	if (prev == NULL) {
		/* link new item into master groups-list */
		master_list->lastlist = gp;
		gp->nextlist = master_list;
		master_list = gp;
	}
	else {
		/* chain into list after 'prev' */
		if (prev->nextgrp) {
			/* I can only chain to the END of a list! */
			fprintf(stderr, "Cannot chain to mid-list!\n");
			return gp;
		}

		gp->lastgrp = prev;
		prev->nextgrp = gp;
	}

	if (channel != -1)
		gp->ch1to4 = newcode(NULL, NULL, index, code);
	else
		gp->rest = newcode(NULL, NULL, index, code);

	return gp;
}


struct group *
newgroup( int type, int id, int len, long tblindex )
{
	struct group	*gp = ( struct group * ) malloc(sizeof( struct group ));

	if (gp) {
		gp->serial_no = group_count++;
		gp->type = type;
		gp->id = id;
		gp->len = len;
		gp->tblindex = tblindex;
		gp->nextgrp = NULL;
		gp->lastgrp = NULL;
		gp->nextlist = NULL;
		gp->lastlist = NULL;
		gp->ch1to4 = NULL;
		gp->rest = NULL;
	}

	return gp;
}


struct code *
newcode( struct code *last, struct code *next, int index, long code)
{
	struct code	*cp = ( struct code * ) malloc(sizeof( struct code ));

	if (cp) {
		cp->serial_no = code_count++;
		cp->index = index;
		cp->code = code;
		cp->next = next;

		if (last)
			last->next = cp;
	}

	return cp;
}

void
write_table( void )
{
	struct group	*gp, *tp;
	struct code	chead, *cp;
	FILE		*out = stdout;

	chead.chain = NULL;
	cp = &chead;

	if (unlink("list7tbl.old") != 0) {
		if (errno != ENOENT) {
			fprintf(stderr, "Cannot remove old table\n");
			return;
		}
	}

	if (rename("list7.tbl", "list7tbl.old") != 0) {
		if (errno != ENOENT) {
			fprintf(stderr, "Cannot rename old table\n");
			return;
		}
	}

	if ((out = fopen("list7.tbl", "w")) == NULL) {
		fprintf(stderr, "Cannot write table\n");
		return;
	}

	fprintf(out, "%d\n", group_count);
	fprintf(out, "%d\n", code_count);
	fprintf(out, "%d\n", (master_list)? master_list->serial_no: -1);
	fprintf(out, "%d\n", known);
	fprintf(out, "%d\n", total);
	fprintf(out, "%d\n", terrestial);
	
	fprintf(out, "#\n# %d groups follow:\n", group_count);
	for (tp = master_list; tp; tp = tp->nextlist) {
		for (gp = tp; gp; gp = gp->nextgrp) {
			fprintf(out, "%d,%d,%d,%d,%ld,%d,%d,%d,%d,%d,%d\n",
				gp->serial_no,
				gp->type,
				gp->id,
				gp->len,
				gp->tblindex,
				((gp->nextgrp)? gp->nextgrp->serial_no: -1),
				((gp->lastgrp)? gp->lastgrp->serial_no: -1),
				((gp->nextlist)? gp->nextlist->serial_no: -1),
				((gp->lastlist)? gp->lastlist->serial_no: -1),
				((gp->ch1to4)? gp->ch1to4->serial_no: -1),
				((gp->rest)? gp->rest->serial_no: -1)
			);
			if (gp->ch1to4) {
				cp = cp->chain = gp->ch1to4;
				while (cp->next)
					cp = cp->chain = cp->next;
			}
			if (gp->rest) {
				cp = cp->chain = gp->rest;
				while (cp->next)
					cp = cp->chain = cp->next;
			}

			cp->chain = NULL;
		}
	}

	fprintf(out, "#\n# %d codes follow:\n", code_count);
	for (cp = chead.chain; cp; cp = cp->chain) {
		fprintf(out, "%d,%ld,%d,%d\n",
				cp->serial_no,
				cp->code,
				cp->index,
				((cp->next)? cp->next->serial_no: -1)
		);
	}

	fclose(out);
	return;
}

char *
fgetl( char *buffer, int bufsiz, FILE *in )
{
	char	*retval;

	do {
		retval = fgets(buffer, bufsiz, in);
	} while (retval != NULL && (buffer[0] == '\n' || buffer[0] == '#'));

	return retval;
}

struct group *
parse_reply( long top5, int first_reqd, struct group *prev, int *np )
{
	static char	buffer[32];
	int		n, start, end, dur, channel;
	int		end_m, end_h;
	int		start_m, start_h, m30, d30;
	long		tblindex;
	struct group	*gp = NULL;

	while (1) {
		fgets(buffer, sizeof buffer, stdin);
		n = sscanf(buffer, "%d %d %d", &channel, &start, &end);
		if (n == 3)
			break;
		if (n == 2) {
			/* channel not given (unknown) */
			end = start;
			start = channel;
			channel = -1;
			break;
		}

		printf("\a?? [ch] start end: ");
	}

	start_h = start / 100;
	start_m = start % 100;
	end_h = end / 100;
	end_m = end % 100;

	if (end_h < start_h)
		end_h += 24;

	dur = (end_h - start_h)*60 + (end_m - start_m);
	tblindex = scanfor(start, dur);

	if (tblindex == -1L || tblindex >= 192L) {
		/* not erratic */
		if (prev == NULL) {
			if (tblindex != -1) {
				printf("\aKnown!\n");
				known++;
			}

			if (channel != -1)
				terrestial++;
		}

		if (((start_m % 30) == 0) && ((dur % 30) == 0)) {
			/* HHB start, HHB duration */
			n = (23 - start_h) * 2;
			d30 = dur;

			if (start_m == 0)
				n++;

			printf("Entry #%d in the 2330->0000/%d HHB group\n", n, dur);
			if (tblindex != -1)
				tblindex -= n;

			gp = insert(HHB, 2330, dur, n, channel, top5, prev, tblindex);

			/* 48 entries in the table */
			if (n == 0) {
				if (!first_reqd)
					printf("\aBugger - missed previous link!\n");
			}
			else if (first_reqd) {
				printf("\aImpossible! - Must be first of group.\n");
				return NULL;
			}

			n = 48 - n;

			if (n == 8) 
				printf("\aBugger - missed next link!\n");
			if (n >= 8)
				/* can't jigsaw next group */
				gp = NULL;
		}
		else if (((start_m % 15) == 0) && ((dur % 30) == 0)) {
			/* QHB start, HHB duration */
			n = start_h * 2;
			d30 = dur;
				
			if (start_m == 45)
				n++;

			printf("Entry #%d in the 0015->2345/%d QHB group\n", n, dur);

			if (tblindex != -1)
				tblindex -= n;

			gp = insert(QHB, 15, dur, n, channel, top5, prev, tblindex);

			/* 48 entries in the table */
			if (n == 0) {
				if (!first_reqd)
					printf("\aBugger - missed previous link!\n");
			}
			else if (first_reqd) {
				printf("\aImpossible! - Must be first of group.\n");
				return NULL;
			}

			n = 48 - n;

			if (n == 8) 
				printf("\aBugger - missed next link!\n");
			if (n >= 8)
				/* can't jigsaw next group */
				gp = NULL;
		}
		else {
			/* assume standard structured group */
			if (start_m >= 30) {
				start_m -= 30;
				m30 = 30;
			}
			else
				m30 = 0;

			if ((dur % 30) == 0) {
				/* must be in the first 4  */
				d30 = dur;
				n = (start_m / 5) - 1;

				if (start_m == 0 || start_m == 15) {
					printf("\aPattern irregularity!\n");
					n = 0;
				}
				else if (n > 2)
					n--;
			}
			else {
				/* the wonders of integer arithmetic */
				d30 = ((dur / 30)+1)*30;
				n = (d30 - dur) / 5;
				n = 4 + 6*(n-1) + (start_m / 5);
			}

			printf("Entry #%d in the %02d%02d->%02d%02d/%d pattern group\n", n, start_h, m30, start_h, m30+25, d30);

			if (tblindex != -1)
				tblindex -= n;

			gp = insert(PATTERN, (start_h*100)+m30, d30, n, channel, top5, prev, tblindex);

			/* 34 entries in the table */
			if (n == 0) {
				if (!first_reqd)
					printf("\aBugger - missed previous link!\n");
			}
			else if (first_reqd) {
				printf("\aImpossible! - Must be first of group.\n");
				return NULL;
			}

			n = 34 - n;

			if (n == 8) 
				printf("\aBugger - missed next link!\n");
			if (n >= 8)
				/* can't jigsaw next group */
				gp = NULL;
		}
	}
	else {
		/* a large erratic group of 192 entries */
		if (prev == NULL)
			printf("\aKnown and Erratic!\n");

		insert(ERRATIC, 0, 0, ( int) tblindex, channel, top5, prev, 0L);
		n = 0;
	}

	*np = n;	/* have to pass this value back too! */
	return gp;
}
