# include	<stdio.h>
# include	<stdlib.h>

/* # include	"vcrplus.h" */
# include	"lookup.h"

static int		start[LIMIT];
static int		leng[LIMIT];
static unsigned char	flags[LIMIT];

# define	GUESS		0x01
# define	SUSPECTDUR	0x02
# define	SUSPECT		0x04
# define	CLASH		0x08
# define	CHANGED		0x80

# define	F_BAK		"lookup.bak"
# ifdef UK
# define	F_TBL		"lookup.uk"
# endif
# ifdef US
# define	F_TBL		"lookup.us"
# endif
# define	F_TMP		"lookup.tmp"

static int	initializer = 0;
static int	change = 0;

static int	tdiff( int a, int b );

void
settables( int start[], int leng[] )
{
	int	j;
	char	buffer[100], *cp;
	FILE	*extras;


/*	for(j=0; j<512; j++) {		 BUG */
	for(j=0; j<LIMIT; j++) {
		start[j] = leng[j] = INVALID;
		flags[j] = 0;
	}

	if (extras = fopen(F_TBL, "r")) {
		while (cp = fgets(buffer, sizeof buffer, extras)) {
			if (*cp == '#' || *cp == '\n')
				/* comment line */
				continue;
	
			if ((j = ( int ) strtol(cp, &cp, 0)) < LIMIT) {
				start[j] = ( int ) strtol(cp, &cp, 0);
				leng[j] = ( int ) strtol(cp, &cp, 0);

				/* scan for flags */
				while (*cp && *cp != '\n') {
					switch (*cp) {
					case 'g':
						flags[j] |= GUESS;
						break;

					case 'D':
						flags[j] |= SUSPECTDUR;
						break;

					case 'S':
						flags[j] |= SUSPECT;
						break;

					case 'c':
						flags[j] |= CLASH;
					}

					cp++;
				}
			}
			else
				printf("Illegal entry %d in lookup.tbl\n", j);
		}

		fclose(extras);
	}
	else {
		fprintf(stderr, "Couldn't open %s\n", F_TBL);
		exit(-1);
	}

	return;
}


/*
 * Confirm: requests 'y/n' confirmation of an entry. Return the actual
 * character typed too.
 */
int
confirm( long index )
{
	char	buffer[20];

	if (flags[index] & GUESS) {
		fputs("\nConfirm this? [y/n] ", stdout);
		fgets(buffer, sizeof buffer, stdin);

		if (buffer[0] == 'y' || buffer[0] == 'Y') {
			flags[index] = CHANGED;
			change = 1;
		}
	}

	return buffer[0];
}



# ifdef OLD
/* insert movie (and other) structured tables */
int tlist[] = { 5, 10, 20, 25,
		0, 5, 10, 15, 20, 25,	/* -5 */
		0, 5, 10, 15, 20, 25,	/* -10 */
		0, 5, 10, 15, 20, 25,	/* -15 */
		0, 5, 10, 15, 20, 25,	/* -20 */
		0, 5, 10, 15, 20, 25	/* -25 */
};

int dlist[] = { 0, 0, 0, 0,
		-5, -5, -5, -5, -5, -5,
		-10, -10, -10, -10, -10, -10,
		-15, -15, -15, -15, -15, -15,
		-20, -20, -20, -20, -20, -20,
		-25, -25, -25, -25, -25, -25
};

static void
fill( int index, int time, int dur)
{
	int	i, s, d;

	for (i = 0; i < 34; i++, index++) {
		s = time + tlist[i];
		d = dur + dlist[i];

		if (start[index] == INVALID) {
			start[index] = s;
			leng[index] = d;
			flags[index] = GUESS | CHANGED;
		}
		else if (start[index] != s || leng[index] != d) {
			/* clash */
			if (flags[index] & GUESS) {
				/* override old guess */
				start[index] = s;
				leng[index] = d;
				flags[index] |= CHANGED;
				printf("Overridden old guess value at %d\n", index);
			}
			else {
				printf("Panic at index=%d\n", index);
				printf("Duff guess: %04d %d\n", s, d);
				exit(-1);
			}
		}
	}

	change = 1;
	return;
}
# endif

long
getguesses( void )
{
	int	BestStart = 0, BestLen = 0;
	int	s, l, i = 0;

	if(initializer == 0) {
		initializer = 1;
		settables(start, leng);
	}

	while (i < SOFTLIMIT) {
		while ((flags[i] & GUESS) == 0 && i < SOFTLIMIT)
			i++;

		/* block starts */
		s = i;
		l = 0;

		while ((flags[i] & GUESS) && i < SOFTLIMIT) {
			/* tot up size of block */
			i++;
			l++;
		}

		if (l > BestLen) {
			BestStart = s;
			BestLen = l;
		}
	}

	printf("Best Start = %d, Best Len = %d\n", BestStart, BestLen);
	return ( long ) BestStart + BestLen/2;
}

void
lookup( long i, int *outtime, int *outdur )
{

	if(initializer == 0) {
		initializer = 1;
		settables(start, leng);
	}

# ifdef DEBUG
	printf("Index used = %d\n", i);
# endif
	
	if((i >= LIMIT) || (i < 0))
		printf("\nPANIC, Illegal table index %ld\n", i);

	if (i >= SOFTLIMIT) {
		/* anything above here is hypothetical anyway */
		*outtime = *outdur = INVALID;
		return;
	}

	fflush(stdout);
	*outtime = start[i];
	*outdur = leng[i];

	if( (*outtime == INVALID) || (*outdur == INVALID) )
	{
		long	k;
		int	tstart, dur;
		char	buffer[32];

		printf("Oops, I don't have an entry for index %ld\a\n", i);
		fputs("Start time? ", stdout);
		fgets(buffer, sizeof buffer, stdin);

		if (buffer[0] >= '0' && buffer[0] <= '9') {
			tstart = atoi(buffer);
			fputs("End time? ", stdout);
			dur = atoi(fgets(buffer, sizeof buffer, stdin));
			dur = tdiff(dur, tstart);

			if ((k = scanfor(tstart, dur)) >= 0) {
				/* clash! */
				printf("Clash with %d\a\n", k);
			}

			start[i] = tstart;
			leng[i] = dur;
			flags[i] = CHANGED;
			change = 1;
		}
		else
			tstart = dur = 0;

		(*outtime) = tstart;
		(*outdur) = dur;
	}
# ifdef DEBUG
	else {
		/* time was OK - how about flags? */
		if (flags[i] & GUESS)
			fputs("This is an educated guess!\n", stdout);
		if (flags[i] & SUSPECTDUR)
			fputs("Duration is suspect - please check\n", stdout);
		if (flags[i] & SUSPECT)
			fputs("Generally suspect - please check\n", stdout);
		if (flags[i] & CLASH)
			fputs("Clash with another entry\n", stdout);
		if (flags[i])
			fputc('\a', stdout);
	}
# endif

	return;
}

# ifdef OLD
static void
pbin( int k )
{
	int	count;

	for (count = 9; count; count--) {
		putchar( (k & 0x100)? '1': '0' );
		k <<= 1;
	}

	return;
}
# endif

long
scanfor( int s, int d )
{
	long	j;

	if (initializer == 0) {
		initializer = 1;
		settables(start, leng);
	}

	for (j = 0; j < SOFTLIMIT; j++) {
		if (start[j] == s && leng[j] == d) {
			/* printf("Match at %d", j); */
			return j;
		}
	}

	return -1L;
}


long
scanforstart( int s, long prev )
{
	long	j;

	if (initializer == 0) {
		initializer = 1;
		settables(start, leng);
	}

	for (j = prev+1; j < SOFTLIMIT; j++) {
		if (start[j] == s) {
			/* printf("Match at %d", j); */
			return j;
		}
	}

	return -1L;
}

int
getduration( long index )
{

	return leng[index];
}

void
dump( void )
{
	static char	buffer[128];
	int		j, k;
	FILE		*in, *out;

	if (initializer == 0)
		/* don't write if we never read! */
		return;

	if (!change)
		return;

	if ((in = fopen(F_TBL, "r")) && (out = fopen(F_TMP, "w"))) {
		j = 0;
		
		while (j < LIMIT) {
			while (j < LIMIT && !(flags[j] & CHANGED))
				j++;
			
			/* found a change - run thru old table file */
			while (fgets(buffer, sizeof buffer, in)) {
				/* comment lines won't trigger this test.. */
				if ((k = atoi(buffer)) >= j)
					break;

				fputs(buffer, out);
			}

			if (feof(in))
				/* special case */
				k = LIMIT;

			/* insert one or more changed lines */
			while (j < LIMIT && j <= k) {
				fprintf(out, "%d\t%d\t%d", j, start[j],
							leng[j]);

				if (flags[j]) {
					fputc('\t', out);

					if (flags[j] & GUESS)
						fputc('g', out);
					if (flags[j] & SUSPECTDUR)
						fputc('D', out);
					if (flags[j] & SUSPECT)
						fputc('S', out);
					if (flags[j] & CLASH)
						fputc('c', out);
				}

				fputc('\n', out);
				j++;

				while (j <= k && !(flags[j] & CHANGED))
					j++;

			}

			if (k < LIMIT && !(flags[k] & CHANGED))
				/* keep line we read in earlier too */
				fputs(buffer, out);
		}

		fclose(out);
		fclose(in);

		/* rename files */
		unlink(F_BAK);
		rename(F_TBL, F_BAK);
		rename(F_TMP, F_TBL);
	}

	return;
}

int
tdiff( int a, int b )
{
	int	hr, min;

	hr = (a / 100) - (b / 100);
	min = (a % 100) - (b % 100);

	if (min < 0) {
		min += 60;
		hr--;
	}

	if (hr < 0)
		hr += 24;

	return (hr*60) + min;
}
